-------------------------------------------------------------------------------
-- Title      : DIOT WIC gateware package
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : wic_pkg.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- This is the package with all types and functions needed by the Warm
-- Interlocks DIOT gateware.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use work.diag_pkg.all;

package wic_pkg is
  
  constant c_IOS_SLOT   : integer := 16;
  constant c_BYTES_SLOT : integer := 4;

  constant c_VAR3_ERR_OFFSET     : integer := 32;
  constant c_VAR3_FAN_ST_OFFSET  : integer := 64;
  constant c_VAR3_TEMP_ST_OFFSET : integer := 65;
  constant c_VAR3_VOLT_ST_OFFSET : integer := 66;
  constant c_VAR3_FANS_OFFSET    : integer := 67;
  constant c_VAR3_TEMP_OFFSET    : integer := 73;

  type t_wic_slot_in is record
    loops : std_logic_vector(c_IOS_SLOT-1 downto 0);
  end record;

  type t_wic_slot_out is record
    relays : std_logic_vector(c_IOS_SLOT-1 downto 0);
  end record;

  constant c_ZERO_SLOT_IN  : t_wic_slot_in := (loops=>(others=>'0'));
  constant c_ZERO_SLOT_OUT : t_wic_slot_out := (relays=>(others=>'0'));

  type t_wic_slot_in_array  is array (natural range <>) of t_wic_slot_in;
  type t_wic_slot_out_array is array (natural range <>) of t_wic_slot_out;

  type t_nf_var is array (natural range <>) of std_logic_vector(7 downto 0);
  type t_wic_var is array (0 to 123) of std_logic_vector(7 downto 0);

  function f_parse_var1 (var : t_wic_var; slot : integer)
    return t_wic_slot_out;

  function f_pack_var3 (slot_array : t_wic_slot_in_array;
    tgl_array : t_wic_slot_in_array; slots : integer; diags : t_diag_out)
    return t_wic_var;

  function f_detect_change (current : t_wic_slot_in_array;
    prev : t_wic_slot_in_array; slots : integer)
    return boolean;

end wic_pkg;

package body wic_pkg is

  -----------------------------------------------------------------------------
  -- function f_pack_var1()
  -- Extracts information from VAR1 for a given slot.
  -----------------------------------------------------------------------------
  function f_parse_var1 (var : t_wic_var; slot : integer)
    return t_wic_slot_out is

    variable tmp : t_wic_slot_out;
    variable i, ui : integer;
  begin
    i  := 0;
    ui := 8;  -- upper index
    tmp.relays := (others=>'0');
    while (i < c_BYTES_SLOT and ui <= c_IOS_SLOT) loop
      tmp.relays(ui-1 downto i*8) := var(slot*c_BYTES_SLOT + i);
      i  := i + 1;
      ui := (i+1)*8;
    end loop;
    return tmp;
  end function;

  -----------------------------------------------------------------------------
  -- function f_pack_var3()
  -- Packs state of input loops into a nanoFIP VAR3 bytes. Takes into account
  -- VAR3 layout - how many bytes per slot should be used.
  -----------------------------------------------------------------------------
  function f_pack_var3 (slot_array : t_wic_slot_in_array;
    tgl_array : t_wic_slot_in_array; slots : integer; diags : t_diag_out)
    return t_wic_var is

    variable i, ui : integer;
    variable sl : integer;
    variable tmp_var : t_wic_var;
  begin
    tmp_var := (others=>(others=>'0'));
    -- iterate through all slots
    for sl in 0 to slots-1 loop
      i  := 0;
      ui := 8;
      while (i < c_BYTES_SLOT and ui <= c_IOS_SLOT) loop
        -- copy input loops state
        tmp_var(sl*c_BYTES_SLOT + i) := slot_array(sl).loops(ui-1 downto i*8);
        -- copy error flags for input loops
        tmp_var(c_VAR3_ERR_OFFSET + sl*c_BYTES_SLOT + i) :=
          tgl_array(sl).loops(ui-1 downto i*8);
        i  := i + 1;
        ui := (i+1)*8;
      end loop;
      -- fill the rest for a given slot with zeros
      -- for both loops state and error flags
      while (i < c_BYTES_SLOT) loop
        tmp_var(sl*c_BYTES_SLOT + i) := (others=>'0');
        tmp_var(c_VAR3_ERR_OFFSET + sl*c_BYTES_SLOT + i) := (others=>'0');
        i := i + 1;
      end loop;
    end loop;
    -- fill monitoring data
    tmp_var(c_VAR3_FAN_ST_OFFSET)  := diags.fan_status;
    tmp_var(c_VAR3_TEMP_ST_OFFSET) := diags.temp_status;
    tmp_var(c_VAR3_VOLT_ST_OFFSET) := diags.volt_status;
    for i in 0 to diags.fans_speed'high loop
      tmp_var(c_VAR3_FANS_OFFSET+i) := diags.fans_speed(i);
    end loop;
    for i in 0 to diags.temps'high loop
      tmp_var(c_VAR3_TEMP_OFFSET+i) := diags.temps(i);
    end loop;
    return tmp_var;
  end function;

  -----------------------------------------------------------------------------
  -- function f_detect_change()
  -- Checks if state of any loop has changed comparing to the value previously
  -- stored.
  -----------------------------------------------------------------------------
  function f_detect_change (current : t_wic_slot_in_array;
    prev : t_wic_slot_in_array; slots : integer)
    return boolean is

    variable i : integer;
    variable diff : boolean;
  begin
    diff := false;
    i    := 0;
    while (i<slots) loop
      if current(i).loops /= prev(i).loops then
        diff := true;
      end if;
      i := i + 1;
    end loop;
    return diff;
  end function;

end wic_pkg;
