-------------------------------------------------------------------------------
-- Title      : DIOT WIC GEFE
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : diot_sb_test_top.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- DIOT top module with DIOT WIC. Done for first gateware tests in hardware,
-- while we are waiting for the first System Board to be produced.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use work.wic_pkg.all;
use work.diag_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity diot_sb_test_top is
  port (
    clk_25m_i : in std_logic;

    -- nanoFIP Wishbone
    dat_i  : in  std_logic_vector(7 downto 0);
    dat_o  : out std_logic_vector(7 downto 0);
    adr_o  : out std_logic_vector(9 downto 0);
    stb_o  : out std_logic;
    wclk_o : out std_logic;
    we_o   : out std_logic;
    cyc_o  : out std_logic;
    ack_i  : in  std_logic;
    rst_o  : out std_logic;
    -- nanoFIP control
    rstin_o    : out std_logic;
    nostat_o   : out std_logic;
    var3_acc_o : out std_logic;
    var3_rdy_i : in  std_logic;
    var2_acc_o : out std_logic;
    var2_rdy_i : in  std_logic;
    var1_acc_o : out std_logic;
    var1_rdy_i : in  std_logic;
    p3_lgth_o  : out std_logic_vector(2 downto 0);

    -- backplane connections
    -- in the first HW demonstrator not all slots are fully populated
    -- crate monitoring
    diag_scl_b  : inout std_logic;
    diag_sda_b  : inout std_logic;
--
--    shared_bus : inout std_logic_vector(3 downto 0);
--    irq_n_i    : in  std_logic;

    s1_p_b  : inout std_logic_vector(25 downto 0);
    s1_n_b  : inout std_logic_vector(25 downto 0);
    s1_en_i : out    std_logic;
    s2_p_b  : inout std_logic_vector(25 downto 0); 
    s2_n_b  : inout std_logic_vector(25 downto 0); 
    s2_en_i : out    std_logic;
    s3_p_b  : inout std_logic_vector(13 downto 0);
    s3_n_b  : inout std_logic_vector(13 downto 0);
    s3_en_i : out    std_logic;
    s4_p_b  : inout std_logic_vector(13 downto 0);
    s4_n_b  : inout std_logic_vector(13 downto 0);
    s4_en_i : out    std_logic;
    s5_p_b  : inout std_logic_vector(13 downto 0);
    s5_n_b  : inout std_logic_vector(13 downto 0);
    s5_en_i : out    std_logic;
    -- slot6 not available in first HW demonstrator
    s7_p_b  : inout std_logic_vector(17 downto 0);
    s7_n_b  : inout std_logic_vector(17 downto 0);
    s7_en_i : out    std_logic;
    -- slot8 not available in first HW demonstrator

    leds_o  : out std_logic_vector(5 downto 0)
  );
end diot_sb_test_top;

architecture rtl of diot_sb_test_top is

  component dummy_clk_gen
    generic (
      g_NUM      : integer := 16;
      g_PORT_NUM : integer := 1;
      g_HI_Z     : boolean := false);
    port (
      rst_n_i : in  std_logic;
      clk_i   : in  std_logic;
      gen_o   : out std_logic_vector(g_NUM-1 downto 0)
    );
  end component;

  constant c_IO_EMULATION : boolean := true;
  constant c_SLOT_NUM     : integer := 6;
  constant c_PERIOD       : unsigned(24 downto 0) := to_unsigned(25000000, 25);

  signal cnt       : unsigned(24 downto 0);
  signal leds      : std_logic_vector(5 downto 0);
  signal rst_cnt   : unsigned(5 downto 0) := (others=> '0');
  signal rst_n     : std_logic;
  signal scl_out, sda_out  : std_logic;
  signal scl_oen, sda_oen  : std_logic;

  signal slots_in  : t_wic_slot_in_array(c_SLOT_NUM-1 downto 0);
  signal slots_out : t_wic_slot_out_array(c_SLOT_NUM-1 downto 0);
  signal diags     : t_diag_out;

  signal s1_temp : std_logic_vector(51 downto 0);
  signal s2_temp : std_logic_vector(43 downto 0);
  signal s3_temp : std_logic_vector(27 downto 0);
  signal s4_temp : std_logic_vector(17 downto 0);
  signal s5_temp : std_logic_vector(27 downto 0);
  signal s7_temp : std_logic_vector(35 downto 0);
  signal s_hi_z  : std_logic_vector(7 downto 0);

begin

  wclk_o <= clk_25m_i;

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if and_reduce(std_logic_vector(rst_cnt)) /= '1' then
        rst_cnt <= rst_cnt + 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when and_reduce(std_logic_vector(rst_cnt))='1' else '0';

-------------------------------------------------------------------------------

  cmp_wic_main: entity work.wic_main
  generic map (
    g_NUM => c_SLOT_NUM)
  port map (
    rst_n_i    => rst_n,
    clk_i      => clk_25m_i,
    slots_i    => slots_in,
    slots_o    => slots_out,

    diags_i    => diags,

    wb_cyc_o   => cyc_o,
    wb_stb_o   => stb_o,
    wb_we_o    => we_o,
    wb_adr_o   => adr_o,
    wb_dat_i   => dat_i,
    wb_dat_o   => dat_o,
    wb_ack_i   => ack_i,
    rst_o      => rst_o,

    rstin_o    => rstin_o,
    nostat_o   => nostat_o,
    var3_acc_o => var3_acc_o,
    var3_rdy_i => var3_rdy_i,
    var2_acc_o => var2_acc_o,
    var2_rdy_i => var2_rdy_i,
    var1_acc_o => var1_acc_o,
    var1_rdy_i => var1_rdy_i,
    p3_lgth_o  => p3_lgth_o);

-------------------------------------------------------------------------------
  
  cmp_diot_diag: entity work.diot_diag
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    scl_o   => scl_out,
    scl_oen => scl_oen,
    scl_i   => diag_scl_b,
    sda_o   => sda_out,
    sda_oen => sda_oen,
    sda_i   => diag_sda_b,
    diags_o => diags);

  --diag_scl_b <= scl_out when scl_oen = '0' else 'Z';
  --diag_sda_b <= sda_out when sda_oen = '0' else 'Z';

-------------------------------------------------------------------------------

GEN_IO_EMULATOR: if c_IO_EMULATION = true generate
  GEN_PORTS: for I in 0 to c_SLOT_NUM-1 generate
    cmp_wic_simulator: entity work.wic_simulator
    generic map (
      g_SIMULATION => true,
      g_ENABLE     => false)
    port map (
      rst_n_i => rst_n,
      clk_i   => clk_25m_i,
      slot_i  => slots_out(I),
      slot_o  => slots_in(I));
  end generate;
end generate;

-------------------------------------------------------------------------------
GEN_REAL_IO: if c_IO_EMULATION = false generate
  -- slots assignment
  slots_in(0).loops              <= s1_p_b(15 downto 0);
  slots_in(1).loops(13 downto 0) <= s2_p_b(13 downto 0);
  slots_in(2).loops(13 downto 0) <= s3_p_b(13 downto 0);
  slots_in(3).loops(8 downto 0)  <= s4_p_b(13 downto 6) & s4_p_b(3);
  slots_in(4).loops(13 downto 0) <= s5_p_b(13 downto 0);
  slots_in(5).loops              <= s7_p_b(15 downto 0);
  s1_n_b(15 downto 0) <= slots_out(0).relays;
  s2_n_b(15 downto 0) <= slots_out(1).relays;
  s3_n_b(13 downto 0) <= slots_out(2).relays(13 downto 0);
  s4_n_b(13 downto 6) <= slots_out(3).relays(8 downto 1);
  s4_n_b(3)           <= slots_out(3).relays(0);
  s5_n_b(13 downto 0) <= slots_out(4).relays(13 downto 0);
  s7_n_b(15 downto 0) <= slots_out(5).relays;

  -- not used loops
  slots_in(1).loops(15 downto 14) <= (others=>'0');
  slots_in(2).loops(15 downto 14) <= (others=>'0');
  slots_in(3).loops(15 downto  9) <= (others=>'0');
  slots_in(4).loops(15 downto 14) <= (others=>'0');
end generate;

-------------------------------------------------------------------------------

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if rst_n = '0' then
        cnt  <= (others=>'0');
        leds <= "000001";
      else
        if cnt = to_unsigned(25000000, 25) then
          cnt <= (others=>'0');
          leds(5 downto 0) <= leds(4 downto 0) & leds(5);
        else
          cnt <= cnt + 1;
        end if;
      end if;
    end if;
  end process;

  GEN_LEDS_O: for I in 0 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;

-------------------------------------------------------------------------------
  s1_p_b <= s1_temp(25 downto 0);
  s1_n_b <= s1_temp(51 downto 26);
  s2_p_b(13 downto 0) <= s2_temp(13 downto 0);
  s2_p_b(25 downto 18)<= s2_temp(21 downto 14);
  s2_n_b(13 downto 0) <= s2_temp(35 downto 22);
  s2_n_b(25 downto 18)<= s2_temp(43 downto 36);
  s3_p_b <= s3_temp(13 downto 0);
  s3_n_b <= s3_temp(27 downto 14);
  s4_p_b(3)           <= s4_temp(0);
  s4_p_b(13 downto 6) <= s4_temp(8 downto 1);
  s4_n_b(3)           <= s4_temp(9);
  s4_n_b(13 downto 6) <= s4_temp(17 downto 10);
  s5_p_b <= s5_temp(13 downto 0);
  s5_n_b <= s5_temp(27 downto 14);
  s7_p_b <= s7_temp(17 downto 0);
  s7_n_b <= s7_temp(35 downto 18);
  diag_scl_b <= s_hi_z(0);
  diag_sda_b <= s_hi_z(1);
  s1_en_i    <= s_hi_z(2);
  s2_en_i    <= s_hi_z(3);
  s3_en_i    <= s_hi_z(4);
  s4_en_i    <= s_hi_z(5);
  s5_en_i    <= s_hi_z(6);
  s7_en_i    <= s_hi_z(7);

  cmd_gen_s1: dummy_clk_gen
  generic map(
    g_NUM      => 52,
    g_PORT_NUM => 1)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s1_temp);

  cmd_gen_s2: dummy_clk_gen
  generic map(
    g_NUM      => 44,
    g_PORT_NUM => 2)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s2_temp);

  cmd_gen_s3: dummy_clk_gen
  generic map(
    g_NUM      => 28,
    g_PORT_NUM => 3)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s3_temp);

  cmd_gen_s4: dummy_clk_gen
  generic map(
    g_NUM      => 18,
    g_PORT_NUM => 4)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s4_temp);

  cmd_gen_s5: dummy_clk_gen
  generic map(
    g_NUM      => 28,
    g_PORT_NUM => 5)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s5_temp);

  cmd_gen_s7: dummy_clk_gen
  generic map(
    g_NUM      => 36,
    g_PORT_NUM => 6)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s7_temp);

  cmd_gen_sz: dummy_clk_gen
  generic map(
    g_NUM      => 8,
    g_PORT_NUM => 7,
    g_HI_Z     => true)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    gen_o   => s_hi_z);

end rtl;
