-------------------------------------------------------------------------------
-- Title      : DIOT WIC Demonstrator
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : diot_wic_top.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- DIOT top module with for Warm Interlocks Controller demo.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the License) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wic_pkg.all;
use work.diag_pkg.all;
use work.wishbone_pkg.all;

entity diot_urv_top is
  port (
    clk_25m_i : in std_logic;
    por_n_b   : inout std_logic;  -- Power-on-reset according to Application note AC380
    button_i  : in std_logic;

    -- nanoFIP Wishbone
    dat_i  : in  std_logic_vector(7 downto 0);
    dat_o  : out std_logic_vector(7 downto 0);
    adr_o  : out std_logic_vector(9 downto 0);
    stb_o  : out std_logic;
    wclk_o : out std_logic;
    we_o   : out std_logic;
    cyc_o  : out std_logic;
    ack_i  : in  std_logic;
    rst_o  : out std_logic;     --  Active high
    -- nanoFIP control
    rstin_o    : out std_logic; --  Active low
    nostat_o   : out std_logic;
    var3_acc_o : out std_logic;
    var3_rdy_i : in  std_logic;
    var2_acc_o : out std_logic;
    var2_rdy_i : in  std_logic;
    var1_acc_o : out std_logic;
    var1_rdy_i : in  std_logic;
    p3_lgth_o  : out std_logic_vector(2 downto 0);

    -- backplane connections
    -- in the first HW demonstrator not all slots are fully populated
    -- crate monitoring
    diag_scl_b  : inout std_logic;
    diag_sda_b  : inout std_logic;
--
--    shared_bus : inout std_logic_vector(3 downto 0);
--    irq_n_i    : in  std_logic;

    s1_p_b  : inout std_logic_vector(15 downto 0);
    s1_n_b  : inout std_logic_vector(15 downto 0);
    s1_en_i : in    std_logic;
    s2_p_b  : inout std_logic_vector(13 downto 0); 
    s2_n_b  : inout std_logic_vector(13 downto 0); 
    s2_en_i : in    std_logic;
    s3_p_b  : inout std_logic_vector(13 downto 0);
    s3_n_b  : inout std_logic_vector(13 downto 0);
    s3_en_i : in    std_logic;
    s4_p_b  : inout std_logic_vector(13 downto 0);
    s4_n_b  : inout std_logic_vector(13 downto 0);
    s4_en_i : in    std_logic;
    s5_p_b  : inout std_logic_vector(13 downto 0);
    s5_n_b  : inout std_logic_vector(13 downto 0);
    s5_en_i : in    std_logic;
    -- slot6 not available in first HW demonstrator
    s7_p_b  : inout std_logic_vector(15 downto 0);
    s7_n_b  : inout std_logic_vector(15 downto 0);
    s7_en_i : in    std_logic;
    -- slot8 not available in first HW demonstrator

    leds_o  : out std_logic_vector(5 downto 0)
  );
end diot_urv_top;

architecture rtl of diot_urv_top is

  component BIBUF_LVCMOS25
  port (
    pad : inout std_logic;
    d   : in    std_logic;
    e   : in    std_logic;
    y   : out   std_logic);
  end component;

  component DFN1C1
  port (
    d   : in  std_logic;
    clk : in  std_logic;
    clr : in  std_logic;
    q   : out std_logic);
  end component;

  constant c_SLOT_NUM     : integer := 6;
  constant c_PERIOD       : unsigned(24 downto 0) := to_unsigned(25000000, 25);

  signal por_n_buf : std_logic;
  signal por_n_d   : std_logic_vector(1 downto 0);
  signal leds      : std_logic_vector(5 downto 0);
  subtype t_rst_cnt is natural range 0 to 15;
  signal rst_cnt   : t_rst_cnt;
  signal rst_n     : std_logic;
  
  signal slots_in        : t_wic_slot_in_array(c_SLOT_NUM-1 downto 0);
  signal slots_out       : t_wic_slot_out_array(c_SLOT_NUM-1 downto 0);
  signal var1_rdy_synced : std_logic;
  signal var3_rdy_synced : std_logic;

  signal wb_fip_in : t_wishbone_slave_in;
  signal wb_fip_out : t_wishbone_slave_out;
  signal sel : std_logic_vector(3 downto 0);
  signal unc_dat : std_logic_vector(31 downto 8);
  signal unc_adr : std_logic_vector(31 downto 12);
begin

-------------------------------------------------------------------------------
-- Power-on-reset according to Microsemi Application Note AC380
-------------------------------------------------------------------------------
  cmp_por_bibuf: BIBUF_LVCMOS25
  port map (
    pad => por_n_b,
    d   => '0',
    e   => '1',
    y   => por_n_buf);

  cmd_por_DFF1: DFN1C1
  port map (
    d   => '1',
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(0));

  cmd_por_DFF2: DFN1C1
  port map (
    d   => por_n_d(0),
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(1));

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if (por_n_d(1) = '0' or button_i = '1') then
        rst_cnt <= t_rst_cnt'high;
      elsif rst_cnt /= 0 then
        rst_cnt <= rst_cnt - 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when rst_cnt = 0 else
           '0';

-------------------------------------------------------------------------------
-- Synchronizers for RDY signals (40MHz clock domain in nanoFIP
-------------------------------------------------------------------------------
  cmp_sync_ffs_var1: entity work.gc_sync
    port map(
      clk_i      => clk_25m_i,
      rst_n_a_i  => rst_n,
      d_i        => var1_rdy_i,
      q_o        => var1_rdy_synced);
  cmp_sync_ffs_var3: entity work.gc_sync
    port map(
      clk_i    => clk_25m_i,
      rst_n_a_i  => rst_n,
      d_i   => var3_rdy_i,
      q_o => var3_rdy_synced);

-------------------------------------------------------------------------------

  inst_fip_urv: entity work.fip_urv
    generic map (
      g_iram_init => "../../../sw/fip_urv/fip_urv.ram"
    )
    port map (
      clk_sys_i => clk_25m_i,
      rst_n_i => rst_n,
      dwb_o => wb_fip_in,
      dwb_i => wb_fip_out
    );

  inst_fip_regs: entity work.fip_urv_regs
    port map (
      rst_n_i => rst_n,
      clk_i => clk_25m_i,
      wb_i => wb_fip_in,
      wb_o => wb_fip_out,
      plc_ctrl_rst_o => open,
      fip_status_var1_rdy_i => var1_rdy_synced,
      fip_status_var3_rdy_i => var3_rdy_synced,
      presence_en_i (0) => s1_en_i,
      presence_en_i (1) => s2_en_i,
      presence_en_i (2) => s3_en_i,
      presence_en_i (3) => s4_en_i,
      presence_en_i (4) => s5_en_i,
      presence_en_i (5) => s7_en_i,
      presence_en_i (6) => '0',
      presence_en_i (7) => '0',
      leds_val_o => leds,
      boards_pins_0_i (15 downto 0) => slots_in(0).loops,
      boards_pins_0_i (31 downto 16) => slots_out(0).relays,
      boards_pins_1_i (15 downto 0) => slots_in(1).loops,
      boards_pins_1_i (31 downto 16) => slots_out(1).relays,
      boards_pins_2_i (15 downto 0) => slots_in(2).loops,
      boards_pins_2_i (31 downto 16) => slots_out(2).relays,
      boards_pins_3_i (15 downto 0) => slots_in(3).loops,
      boards_pins_3_i (31 downto 16) => slots_out(3).relays,
      boards_pins_4_i (15 downto 0) => slots_in(4).loops,
      boards_pins_4_i (31 downto 16) => slots_out(4).relays,
      boards_pins_5_i (15 downto 0) => slots_in(5).loops,
      boards_pins_5_i (31 downto 16) => slots_out(5).relays,
      boards_pins_6_i => x"00_00_00_00",
      boards_pins_7_i => x"00_00_00_00",
      fip_reg_i.dat (7 downto 0) => dat_i,
      fip_reg_i.dat (31 downto 8) => x"00_00_00",
      fip_reg_i.ack => ack_i,
      fip_reg_i.err => '0',
      fip_reg_i.rty => '0',
      fip_reg_i.stall => '0',
      fip_reg_o.dat (7 downto 0) => dat_o,
      fip_reg_o.dat (31 downto 8) => unc_dat,
      fip_reg_o.adr (11 downto 2) => adr_o,
      fip_reg_o.adr (31 downto 12) => unc_adr,
      fip_reg_o.cyc => cyc_o,
      fip_reg_o.stb => stb_o,
      fip_reg_o.sel => sel,
      fip_reg_o.we => we_o,
      plc_mem_i.ack => '1',
      plc_mem_i.dat => x"00_00_00_00",
      plc_mem_i.err => '0',
      plc_mem_i.rty => '0',
      plc_mem_i.stall => '0',
      plc_mem_o => open
    );

  wclk_o <= clk_25m_i;
  rst_o <= not rst_n;
  nostat_o <= '1';
  rstin_o <= rst_n;

  var3_acc_o <= '0';
  var2_acc_o <= '0';
  var1_acc_o <= '0';
  p3_lgth_o  <= "001"; -- 8 bytes

  diag_scl_b <= 'Z';
  diag_sda_b <= 'Z';

-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
  GEN_REAL_IO: block
  begin
    -- slots assignment
    slots_in(0).loops              <= s1_p_b(15 downto 0);
    slots_in(1).loops(13 downto 0) <= s2_p_b(13 downto 0);
    slots_in(2).loops(13 downto 0) <= s3_p_b(13 downto 0);
    slots_in(3).loops(8 downto 0)  <= s4_p_b(13 downto 6) & s4_p_b(3);
    slots_in(4).loops(13 downto 0) <= s5_p_b(13 downto 0);
    slots_in(5).loops              <= s7_p_b(15 downto 0);
    s1_n_b(15 downto 0) <= slots_out(0).relays;
    s2_n_b(13 downto 0) <= slots_out(1).relays(13 downto 0);
    s3_n_b(13 downto 0) <= slots_out(2).relays(13 downto 0);
    s4_n_b(13 downto 6) <= slots_out(3).relays(8 downto 1);
    s4_n_b(3)           <= slots_out(3).relays(0);
    s5_n_b(13 downto 0) <= slots_out(4).relays(13 downto 0);
    s7_n_b(15 downto 0) <= slots_out(5).relays;
  
    -- not used loops
    slots_in(1).loops(15 downto 14) <= (others=>'0');
    slots_in(2).loops(15 downto 14) <= (others=>'0');
    slots_in(3).loops(15 downto  9) <= (others=>'0');
    slots_in(4).loops(15 downto 14) <= (others=>'0');
  end block;

  --  Leds: assign to '0' to switch on the led, 'Z' to switch off.
  leds_o(0) <= '0' when rst_n = '0' or leds(0) = '1' else '1';

  GEN_LEDS_O: for I in 1 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;
end rtl;
