-------------------------------------------------------------------------------
-- Title      : DIOT WIC GEFE
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : sb_fakeio_top.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- DIOT top module with DIOT WIC. Done for first gateware tests in hardware,
-- while we are waiting for the first System Board to be produced.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use work.wic_pkg.all;

entity sb_fakeio_top is
  port (
    clk_25m_i : in std_logic;

    -- backplane connections
    -- in the first HW demonstrator not all slots are fully populated
--    i2c_scl_o  : out   std_logic;
--    i2c_sda_b  : inout std_logic;
--
--    shared_bus : inout std_logic_vector(3 downto 0);
--    irq_n_i    : in  std_logic;

    slot_p_o  : out std_logic_vector(25 downto 0);
    slot_n_i  : in  std_logic_vector(25 downto 0);

    leds_o  : out std_logic_vector(5 downto 0)
  );
end sb_fakeio_top;

architecture rtl of sb_fakeio_top is

  constant c_PERIOD       : unsigned(24 downto 0) := to_unsigned(10000000, 25);

  signal cnt       : unsigned(24 downto 0);
  signal leds      : std_logic_vector(5 downto 0);
  signal rst_cnt   : unsigned(5 downto 0) := (others=> '0');
  signal rst_n     : std_logic;
  signal slot_n_synced : std_logic_vector(c_IOS_SLOT-1 downto 0);
  signal slots_in  : t_wic_slot_in;
  signal slots_out : t_wic_slot_out;

begin

-------------------------------------------------------------------------------

  GEN_LOOPS_SYNC: for I in 0 to c_IOS_SLOT-1 generate
    cmp_sync_ffs: entity work.gc_sync_ffs
      port map(
        clk_i    => clk_25m_i,
        rst_n_i  => rst_n,
        data_i   => slot_n_i(I),
        synced_o => slot_n_synced(I),
        npulse_o => open,
        ppulse_o => open);
  end generate;

-------------------------------------------------------------------------------

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if and_reduce(std_logic_vector(rst_cnt)) /= '1' then
        rst_cnt <= rst_cnt + 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when and_reduce(std_logic_vector(rst_cnt))='1' else '0';

-------------------------------------------------------------------------------

  cmp_wic_simulator: entity work.wic_simulator
  generic map (
    g_SIMULATION => true,
    g_ENABLE     => false)
  port map (
    rst_n_i => rst_n,
    clk_i   => clk_25m_i,
    slot_i  => slots_out,
    slot_o  => slots_in);

-------------------------------------------------------------------------------
  -- slots assignment
  slot_p_o(15 downto 0) <= slots_in.loops;
  slots_out.relays  <= slot_n_synced(15 downto 0);
  slot_p_o(25 downto 16) <= (others=>'0');

-------------------------------------------------------------------------------

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if rst_n = '0' then
        cnt  <= (others=>'0');
        leds <= "000001";
      else
        if cnt = c_PERIOD then
          cnt <= (others=>'0');
          leds(5 downto 0) <= leds(4 downto 0) & leds(5);
        else
          cnt <= cnt + 1;
        end if;
      end if;
    end if;
  end process;

  GEN_LEDS_O: for I in 0 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;

-------------------------------------------------------------------------------

end rtl;
