-------------------------------------------------------------------------------
-- Title      : DIOT WIC SPEC
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : spec_top.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- SPEC top module with DIOT WIC. Done for first gateware tests in hardware,
-- while we are waiting for the first System Board to be produced.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use work.wic_pkg.all;
use work.diag_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity spec_top is
  port (
    clk_125m_pllref_p_i : in std_logic;
    clk_125m_pllref_n_i : in std_logic;

    -- nanoFIP Wishbone
    dat_i  : in  std_logic_vector(7 downto 0);
    dat_o  : out std_logic_vector(7 downto 0);
    adr_o  : out std_logic_vector(9 downto 0);
    stb_o  : out std_logic;
    wclk_o : out std_logic;
    we_o   : out std_logic;
    cyc_o  : out std_logic;
    ack_i  : in  std_logic;
    rst_o  : out std_logic;
    -- nanoFIP control
    rstin_o    : out std_logic;
    nostat_o   : out std_logic;
    var3_acc_o : out std_logic;
    var3_rdy_i : in  std_logic;
    var2_acc_o : out std_logic;
    var2_rdy_i : in  std_logic;
    var1_acc_o : out std_logic;
    var1_rdy_i : in  std_logic;
    p3_lgth_o  : out std_logic_vector(2 downto 0);

    -- crate monitoring
    diag_scl_b : inout std_logic := '1';
    diag_sda_b : inout std_logic := '1'
  );
end spec_top;

architecture rtl of spec_top is

  constant c_SLOT_NUM : integer := 2;

  signal rst_cnt : unsigned(5 downto 0) := (others=> '0');

  signal clk_125m          : std_logic;
  signal clkfb, clkfb_bufg : std_logic;
  signal dcm_clk_out       : std_logic;
  signal dcm_clk_out_n     : std_logic;
  signal clk_sys_25m       : std_logic;
  signal clk_sys_25m_n     : std_logic;
  signal clk_sys_locked    : std_logic;
  signal rst_n             : std_logic;
  signal scl_out, sda_out  : std_logic;
  signal scl_oen, sda_oen  : std_logic;

  signal slots_in  : t_wic_slot_in_array(c_SLOT_NUM-1 downto 0);
  signal slots_out : t_wic_slot_out_array(c_SLOT_NUM-1 downto 0);
  signal diags     : t_diag_out;

  signal s1_temp : std_logic_vector(15 downto 0);
  signal s2_temp : std_logic_vector(20 downto 0);
  signal s8_temp : std_logic_vector(20 downto 0);

begin

-------------------------------------------------------------------------------
-- SPEC board/Xilinx specific components --
-------------------------------------------------------------------------------
  cmp_clk_in: IBUFGDS
  port map (
    O  => clk_125m,
    I  => clk_125m_pllref_p_i,
    IB => clk_125m_pllref_n_i);

  clkf_bufg: BUFG
  port map (
    O => clkfb_bufg,
    I => clkfb); 

  clksys_bufg: BUFG
  port map (
    O => clk_sys_25m,
    I => dcm_clk_out);

  clksys_n_bufg: BUFG
  port map (
    O => clk_sys_25m_n,
    I => dcm_clk_out_n);

  cmp_clk_sys_gen: DCM_SP
  generic map
   (CLKDV_DIVIDE          => 5.000,
    CLKFX_DIVIDE          => 10,
    CLKFX_MULTIPLY        => 2,
    CLKIN_DIVIDE_BY_2     => FALSE,
    CLKIN_PERIOD          => 8.0,
    CLKOUT_PHASE_SHIFT    => "NONE",
    CLK_FEEDBACK          => "1X",
    DESKEW_ADJUST         => "SYSTEM_SYNCHRONOUS",
    PHASE_SHIFT           => 0,
    STARTUP_WAIT          => FALSE)
  port map
   -- Input clock
   (CLKIN                 => clk_125m,
    CLKFB                 => clkfb_bufg,
    -- Output clocks
    CLK0                  => clkfb,
    CLKFX                 => dcm_clk_out,
    CLKFX180              => dcm_clk_out_n,
   -- Ports for dynamic phase shift
    PSCLK                 => '0',
    PSEN                  => '0',
    PSINCDEC              => '0',
   -- Other control and status signals
    LOCKED                => clk_sys_locked,
    RST                   => '0',
    DSSEN                 => '0');

  wclk_oddr: ODDR2
    port map (
        Q       => wclk_o,
        C0      => clk_sys_25m,
        C1      => clk_sys_25m_n,
        CE      => '1',
        D0      => '1',
        D1      => '0',
        R       => '0',
        S       => '0'
    );

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

  process(clk_sys_25m)
  begin
    if rising_edge(clk_sys_25m) then
      if and_reduce(std_logic_vector(rst_cnt)) /= '1' then
        rst_cnt <= rst_cnt + 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when (and_reduce(std_logic_vector(rst_cnt))='1' and clk_sys_locked = '1') else '0';

-------------------------------------------------------------------------------

  cmp_wic_main: entity work.wic_main
  generic map (
    g_NUM => c_SLOT_NUM)
  port map (
    rst_n_i    => rst_n,
    clk_i      => clk_sys_25m,
    slots_i    => slots_in,
    slots_o    => slots_out,

    diags_i    => diags,

    wb_cyc_o   => cyc_o,
    wb_stb_o   => stb_o,
    wb_we_o    => we_o,
    wb_adr_o   => adr_o,
    wb_dat_i   => dat_i,
    wb_dat_o   => dat_o,
    wb_ack_i   => ack_i,
    rst_o      => rst_o,

    rstin_o    => rstin_o,
    nostat_o   => nostat_o,
    var3_acc_o => var3_acc_o,
    var3_rdy_i => var3_rdy_i,
    var2_acc_o => var2_acc_o,
    var2_rdy_i => var2_rdy_i,
    var1_acc_o => var1_acc_o,
    var1_rdy_i => var1_rdy_i,
    p3_lgth_o  => p3_lgth_o);

-------------------------------------------------------------------------------
  
  cmp_diot_diag: entity work.diot_diag
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_sys_25m,
    scl_o   => scl_out,
    scl_oen => scl_oen,
    scl_i   => diag_scl_b,
    sda_o   => sda_out,
    sda_oen => sda_oen,
    sda_i   => diag_sda_b,
    diags_o => diags);

  diag_scl_b <= scl_out when scl_oen = '0' else 'Z';
  diag_sda_b <= sda_out when sda_oen = '0' else 'Z';

-------------------------------------------------------------------------------

  GEN_SIMULATOR: for I in 0 to c_SLOT_NUM-1 generate
    cmp_wic_simulator: entity work.wic_simulator
    generic map (
      g_SIMULATION => false,
      g_ENABLE     => true)
    port map (
      rst_n_i => rst_n,
      clk_i   => clk_sys_25m,
      slot_i  => slots_out(I),
      slot_o  => slots_in(I));
  end generate;

-------------------------------------------------------------------------------

  cmp_gen_s1: entity work.dummy_clk_gen
  generic map(
    g_NUM      => 16,
    g_PORT_NUM => 1)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_sys_25m,
    gen_o   => s1_temp);

  cmp_gen_s2: entity work.dummy_clk_gen
  generic map(
    g_NUM      => 21,
    g_PORT_NUM => 2)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_sys_25m,
    gen_o   => s2_temp);

  cmp_gen_s8: entity work.dummy_clk_gen
  generic map(
    g_NUM      => 21,
    g_PORT_NUM => 8,
    g_HI_Z     => true)
  port map(
    rst_n_i => rst_n,
    clk_i   => clk_sys_25m,
    gen_o   => s8_temp);

end rtl;
