/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "diot_crate.h"
#include <limits>

void DIOT_Crate::SetOutput(int slot, int idx, bool state)
{
    const int var_offset = getDataOffset(slot, idx);
    const uint8_t io_mask = getIOMask(idx);
    uint8_t data = GetProdVarData(var_offset);

    if ((data & io_mask) == state)
        return;     // no update needed

    if (state)
        data |= io_mask;
    else
        data &= ~io_mask;

    SetProdVarData(var_offset, data);
}


void DIOT_Crate::SetSlotOutputs(int slot, uint32_t state)
{
    const int var_offset = getDataOffset(slot, 0);

    {   // mutex locked scope
    auto lock = getProdLock();
    setProdDataLock(var_offset, (uint8_t)(state & 0xff), lock);
    setProdDataLock(var_offset+1, (uint8_t)(state >> 8 & 0xff), lock);
    setProdDataLock(var_offset+2, (uint8_t)(state >> 16 & 0xff), lock);
    setProdDataLock(var_offset+3, (uint8_t)(state >> 24 & 0xff), lock);
    }
}


bool DIOT_Crate::GetFanStatus(unsigned int fan)
{
    if (fan >= FAN_COUNT)
        return false;

    int status = GetFanStatusAll();
    // status bit == 1 indicates failure, hence negation
    return !(status & (1 << fan));
}


int DIOT_Crate::GetFanRPM(unsigned int fan)
{
    if (fan >= FAN_COUNT)
        return std::numeric_limits<int>::min();

    int rpm = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    rpm = (int) getConsDataLock(FAN0_SPEED + fan * 2, lock) << 8;
    rpm |= getConsDataLock(FAN0_SPEED + fan * 2 + 1, lock);
    }

    return rpm;
}


bool DIOT_Crate::GetTempStatus(unsigned int sensor)
{
    if (sensor >= TEMP_COUNT)
        return false;

    int status = GetTempStatusAll();
    // status bit == 1 indicates failure, hence negation
    return !(status & (1 << sensor));
}


int DIOT_Crate::GetTemp(unsigned int sensor)
{
    if (sensor >= TEMP_COUNT)
        return std::numeric_limits<int>::min();

    return GetConsVarData(TEMPERATURE0 + sensor);
}


std::string DIOT_Crate::GetVoltName(VOLTAGE volt)
{
    for (const auto& volt_pair : VOLTAGES) {
        if (volt_pair.first == volt)
            return volt_pair.second;
    }

    return "UNKNOWN";
}


DIOT_Crate::VOLTAGE DIOT_Crate::GetVoltEnum(unsigned int idx)
{
    if (idx >= VOLTAGES.size())
        return INVALID;

    return VOLTAGES[idx].first;
}


uint32_t DIOT_Crate::getConsWord(int address)
{
    uint32_t ret = 0;

    {   // mutex locked scope
    auto lock = getConsLock();

    for (int i = 3; i >= 0; --i) {
        ret <<= 8;
        ret |= getConsDataLock(address + i, lock);
    }
    }

    return ret;
}


const std::vector<std::pair<DIOT_Crate::VOLTAGE, std::string>> DIOT_Crate::VOLTAGES =
    { { P3V3, "+3.3V" }, { P5V0, "+5V" },
      { P12V0, "+12V" }, { M12V0, "-12V" } };
