/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string>
#include <stdexcept>
#include <signal.h>

#include "libmfipp.h"

MasterFIP::MasterFIP(struct mstrfip_dev *device)
    : dev(device), configured(false), state(STOPPED), orig_err_handler(nullptr)
{
    if (!dev) {
        throw std::runtime_error("Null MasterFIP device");
    }

    // First instance needs to initialize MasterFIP library
    if (instances.empty() && mstrfip_init()) {
        throw std::runtime_error(errStr("Cannot init mfip library: "));
    }

    // Restart the RT Mockturtle application
    if (mstrfip_rtapp_reset(dev) < 0) {
        throw std::runtime_error(errStr("Cannot reset rt app: "));
    }

    // Obtain the FIP bus speed
    if (mstrfip_hw_speed_get(dev, &fbus_speed)) {
        throw std::runtime_error(errStr("Failed to get FIP bus speed: "));
    }

    if (instances.count(dev)) {
        throw std::runtime_error("Duplicate mstrfip_dev");
    }

    ClearErrors();
    instances[dev] = this;
}


MasterFIP::~MasterFIP()
{
    if (dev)
        mstrfip_close(dev);

    instances.erase(dev);

    if (instances.empty())
        mstrfip_exit();
}


std::unique_ptr<MasterFIP> MasterFIP::OpenByLUN(uint32_t lun)
{
    std::unique_ptr<MasterFIP> ret(new MasterFIP(mstrfip_open_by_lun(lun)));
    return ret;
}


std::unique_ptr<MasterFIP> MasterFIP::OpenByFMC(uint32_t fmc)
{
    std::unique_ptr<MasterFIP> ret(new MasterFIP(mstrfip_open_by_fmc(fmc)));
    return ret;
}


MasterFIP* MasterFIP::FindByDev(mstrfip_dev *dev)
{
    auto it = instances.find(dev);
    return it == instances.end() ? nullptr : it->second;
}


void MasterFIP::Configure(mstrfip_hw_cfg hw_cfg, mstrfip_sw_cfg sw_cfg)
{
    if (mstrfip_hw_cfg_set(dev, &hw_cfg) < 0) {
        throw std::runtime_error(errStr("Cannot configure HW: "));
    }

    // Replace the error handler to count them,
    // but save the original handler so it is also called
    orig_err_handler = sw_cfg.mstrfip_error_handler;
    sw_cfg.mstrfip_error_handler = &MasterFIP::error_handler;

    if (mstrfip_sw_cfg_set(dev, &sw_cfg) < 0) {
        throw std::runtime_error(errStr("Cannot configure SW: "));
    }

    configured = true;
}


void MasterFIP::StartCycle(CycleBase &cycle)
{
    if (!configured) {
        throw std::runtime_error("MasterFIP has not been configured");
    }

    if (!cycle.IsValid()) {
        throw std::runtime_error("Invalid cycle configuration");
    }

    if (state == RUNNING)
        StopCycle();

    cycle.Update();

    if (mstrfip_ba_load(dev, cycle.GetMacroCycle()) < 0) {
        throw std::runtime_error(errStr("Loading macrocycle failed: "));
    }

    cycle.ReadRespTime();

    if (mstrfip_ba_start(dev) < 0) {
        throw std::runtime_error(errStr("Start BA failed: "));
    }

    state = RUNNING;
}


void MasterFIP::StopCycle()
{
    if(state == STOPPED) {
        DBG("No cycle to stop\n");
        return;
    }

    if (mstrfip_ba_stop(dev)) {
        throw std::runtime_error(errStr("Stop BA failed: "));
    }

    state = STOPPED;
}


void MasterFIP::ClearErrors()
{
    errors[MSTRFIP_INVALID_ANSWER_ACK] = 0;
    errors[MSTRFIP_INVALID_BITRATE] = 0;
    errors[MSTRFIP_INVALID_BA_INSTR_COUNT] = 0;
    errors[MSTRFIP_BA_IS_NULL] = 0;
    errors[MSTRFIP_BA_MAX_VAR] = 0;
    errors[MSTRFIP_BA_INVALID_VAR_COUNT] = 0;
    errors[MSTRFIP_BA_PER_WIND_CFG_BAD] = 0;
    errors[MSTRFIP_BA_INVALID_INSTR_PARAM] = 0;
    errors[MSTRFIP_BA_NOT_STOP_STATE] = 0;
    errors[MSTRFIP_BA_INCONSISTENCY] = 0;
    errors[MSTRFIP_BA_WRONG_CYCLE_LENGTH] = 0;
    errors[MSTRFIP_BA_INCONSISTENT_INSTR_SET] = 0;
    errors[MSTRFIP_BA_WAIT_WIND_INSTR_MISSING] = 0;
    errors[MSTRFIP_BA_INVALID_MACROCYCLE] = 0;
    errors[MSTRFIP_INVALID_HW_VERS] = 0;
    errors[MSTRFIP_TRTL_MSG_COUNT_INCONSISTENCY] = 0;
    errors[MSTRFIP_TRTL_MSG_COUNT_OVERFLOW] = 0;
    errors[MSTRFIP_TRTL_MSG_OVERFLOW] = 0;
    errors[MSTRFIP_TRTL_MSG_READ_ERR] = 0;
    errors[MSTRFIP_TRTL_MSG_READ_NULL] = 0;
    errors[MSTRFIP_POLL_TIMEOUT] = 0;
    errors[MSTRFIP_POLL_ERROR] = 0;
    errors[MSTRFIP_BA_APER_VAR_NO_CB] = 0;
    errors[MSTRFIP_BA_APER_MSG_NO_CB] = 0;
    errors[MSTRFIP_BA_PER_VAR_NO_CB] = 0;
    errors[MSTRFIP_RESET_MOCKTRTL_ERR] = 0;
    // initialize the sentinel, so there are no exceptions when it is accessed
    errors[__MSTRFIP_MAX_ERROR_CODE] = 0;
}


void MasterFIP::error_handler(mstrfip_dev *dev, mstrfip_error_list error)
{
    auto it = MasterFIP::instances.find(dev);

    if (it == MasterFIP::instances.end()) {
        DBG("error_handler could not find 0x%p device\n", dev);
        return;
    }

    MasterFIP* mfip = it->second;
    ++mfip->errors[error];

    if (mfip->orig_err_handler)
        (*mfip->orig_err_handler)(dev, error);
}


std::string MasterFIP::errStr(const std::string& msg)
{
    std::string error_string(msg);
    error_string += mstrfip_strerror(errno);
    return error_string;
}


std::unordered_map<mstrfip_dev*, MasterFIP*> MasterFIP::instances;


WindowBase::~WindowBase()
{
}


WindowVarBase::~WindowVarBase()
{
}


int WindowVarBase::AddVariable(int id, mstrfip_data_flags flags, int max_bsz,
        VarMsgHandlerFn handler)
{
    mstrfip_data* data = doAddVar(id, flags, max_bsz,
            &WindowVarBase::handler_wrapper);

    if(!data)
        return -1;

    if (handler)
        handler_map.insert(std::make_pair(data, handler));

    return 0;
}


mstrfip_data* WindowVarBase::doAddVar(int id, mstrfip_data_flags flags,
        int max_bsz, VarMsgHandlerPtr handler)
{
    mstrfip_data_cfg cfg;
    cfg.id = id;
    cfg.flags = flags;
    cfg.max_bsz = max_bsz;
    cfg.mstrfip_data_handler = handler;
    mstrfip_data* var_data = mstrfip_var_create(cycle.GetMacroCycle(), &cfg);

    if (var_data)
        data.push_back(var_data);

    return var_data;
}


void WindowVarBase::handler_wrapper(mstrfip_dev *dev, mstrfip_data *var,
        mstrfip_irq *irq)
{
    auto it = handler_map.find(var);

    if (it != handler_map.end()) {
        (it->second)(dev, var, irq);
    }
}


std::unordered_map<mstrfip_data*, VarMsgHandlerFn> WindowVarBase::handler_map;


int WindowPeriodicVar::Instantiate()
{
    if (data.empty())
        return -1;

    struct mstrfip_per_var_wind_cfg pwind_cfg = {0};
    pwind_cfg.varlist = data.data();        // looks silly, doesn't it?
    pwind_cfg.var_count = data.size();
    return mstrfip_per_var_wind_append(cycle.GetMacroCycle(), &pwind_cfg);
}


int WindowAperiodicVar::AddIdent(uint32_t agent_addr)
{
    mstrfip_data* var_data = mstrfip_ident_var_create(cycle.GetMacroCycle(),
            agent_addr);

    if (!var_data)
        return -1;

    data.push_back(var_data);
    return 0;
}


int WindowAperiodicVar::Instantiate()
{
    if (data.empty())
        return -1;

    if (end_time <= 0)
        return -1;

    struct mstrfip_aper_var_wind_cfg apwind_cfg = {0};
    apwind_cfg.end_ustime = end_time;
    apwind_cfg.enable_diag = diag ? 1 : 0;
    /* identification variable list that can be requested at run time */
    apwind_cfg.ident_varlist = data.data();
    apwind_cfg.ident_var_count = data.size();
    /* user handler calls back when an ident request has been fully served */
    apwind_cfg.mstrfip_ident_var_handler = callback;
    return mstrfip_aper_var_wind_append(cycle.GetMacroCycle(), &apwind_cfg);
}


int WindowAperiodicMsg::Instantiate()
{
    if (prod_fifo_size < 0 || cons_fifo_size < 0)
        return -1;

    if (prod_fifo_size <= 0 && cons_fifo_size <= 0)
        return -1;

    if (end_time <= 0)
        return -1;

    mstrfip_aper_msg_wind_cfg apwind_cfg = {0};
    apwind_cfg.end_ustime = end_time;
    apwind_cfg.prod_msg_fifo_size = prod_fifo_size;
    apwind_cfg.cons_msg_fifo_size = cons_fifo_size;
    apwind_cfg.mstrfip_cons_msg_handler = cons_cb;
    apwind_cfg.mstrfip_prod_msg_handler = prod_cb;
    return mstrfip_aper_msg_wind_append(cycle.GetMacroCycle(), &apwind_cfg);
}


CycleBase::CycleBase(const MasterFIP &mfip, int length_us)
    : dev(mfip.GetDevice()), length(length_us)
{
    if(length_us <= 0)
        throw std::runtime_error("Cycle length must be positive");

    mcycle = mstrfip_macrocycle_create(dev);

    if (!mcycle)
        throw std::runtime_error("Could not crate macrocycle");
}


CycleBase::~CycleBase()
{
    if (mstrfip_macrocycle_delete(dev, mcycle))
        DBG("Could not delete cycle\n");
}


int CycleBase::ReadRespTime()
{
    std::vector<int> addresses;

    for (const AgentBase* agent : agents) {
        addresses.push_back(agent->GetAddress());
    }

    std::vector<uint32_t> resp_time;
    resp_time.resize(agents.size());

    if (mstrfip_hw_response_time_get(dev, agents.size(),
                addresses.data(), resp_time.data())) {
        DBG(MasterFIP::errStr("Read turn around time failed: "));
        return -1;
    }

    return 0;
}


CycleSimple::CycleSimple(const MasterFIP &mfip, int length_us)
    : CycleBase(mfip, length_us),
    win_per_var(*this), win_aper_var(*this), instantiated(false)
{
}


int CycleSimple::Update()
{
    if (!instantiated) {
        bool empty = true;

        if (!win_per_var.Empty()) {
            win_per_var.Instantiate();
            empty = false;
        }

        if (!win_aper_var.Empty()) {
            // end the aperiodic window at 10% of the cycle
            win_aper_var.SetEnd(length * 9 / 10);
            win_aper_var.Instantiate();
            empty = false;
        }

        if (empty) {
            DBG("Empty macrocyle\n");
            return -1;
        }

        // append a wait window without padding (silent) 
        int res = mstrfip_wait_wind_append(mcycle, 1, length);

        if (res)
            return res;

        instantiated = true;
    }

    return 0;
}

AgentBase::AgentBase(unsigned int addr)
    : address(addr)
{
    if (addr >= 255)
        throw std::runtime_error("Agent address must lower than 255");

    ClearErrors();
}


AgentBase::~AgentBase()
{
}


void AgentBase::ClearErrors()
{
    errors[WRITE_FAILED] = 0;
    errors[MISSED] = 0;
    errors[NOT_FRESHED] = 0;
    errors[NOT_SIGNIFICANT] = 0;
    errors[TIMEOUT] = 0;
    errors[HW_ERR] = 0;
    errors[HW_BSZ] = 0;
    errors[HW_CTRL] = 0;
    errors[HW_PDU] = 0;
    errors[UNKNOWN] = 0;
}


int WaitForSignal()
{
    int signo;
    siginfo_t siginfo;
    sigset_t newmask;

    // run for ever till a signal is raised
    sigemptyset(&newmask); // unblock any signal
    sigaddset(&newmask, SIGINT);
    sigaddset(&newmask, SIGTERM);
    sigaddset(&newmask, SIGABRT);
    sigaddset(&newmask, SIGFPE);
    sigaddset(&newmask, SIGILL);
    sigaddset(&newmask, SIGUSR1);
    pthread_sigmask(SIG_BLOCK, &newmask, NULL);

    while (true) {
        signo = sigwaitinfo(&newmask, &siginfo);

        if (signo > 0)
            return signo;
    }

    return -1;
}
